#include "topo.h"

#define SOUTH	0	/* important that these 4 come before NE,SW,NW,SE */
#define WEST	1
#define NORTH	2
#define EAST	3
#define NW	4	/* important that low bits alternate SW,NE and NW,SE */
#define SE	5
#define NE	6
#define SW	7

/* SOUTH,WEST,NORTH,EAST,NW,SE,NE,SW */
static int lorient[8] = {SE,SW,NW,NE,WEST,EAST,NORTH,SOUTH};
static int rorient[8] = {SW,NW,NE,SE,NORTH,SOUTH,EAST,WEST};
static int b1orient[8] = {EAST,SOUTH,WEST,NORTH,SW,NE,NW,SE};
static int b2orient[8] = {WEST,NORTH,EAST,SOUTH,NE,SW,SE,NW};
static int b3orient[8] = {NORTH,EAST,SOUTH,WEST,SE,NW,SW,NE};

/* NW,SE,NE,SW (parity of x determines row) */
static int porient1[8] = {WEST,SOUTH,NORTH,WEST,
			  NORTH,EAST,EAST,SOUTH};
/* SOUTH,WEST,NORTH,EAST (parity of x+y determines row) */
static int porient0[8] = {SE,NW,NW,SE,
			  SW,SW,NE,NE};

/* SOUTH,WEST,NORTH,EAST,NW,SE,NE,SW */
static int lxinc[8] = {0,0,1,1,0,0,0,0};
static int lyinc[8] = {0,1,1,0,0,0,0,0};
static int rxinc[8] = {1,0,0,1,0,0,0,0};
static int ryinc[8] = {0,0,1,1,0,0,0,0};
static int b1xinc[8] = {0,0,0,0,0,0,+1,-1};
static int b1yinc[8] = {0,0,0,0,+1,-1,0,0};
static int b2xinc[8] = {0,0,0,0,-1,+1,0,0};
static int b2yinc[8] = {0,0,0,0,0,0,+1,-1};
static int b3xinc[8] = {0,-1,0,+1,0,0,0,0};
static int b3yinc[8] = {-1,0,+1,0,0,0,0,0};

static int xmax[64],ymax[64],ntriabove[64];

#define S1(t) 2*(t->y*xmax[t->depth]+t->y+t->x)+(t->orient&1)+ntriabove[t->depth]
#define S0(t) 4*(t->y*xmax[t->depth]+t->y+t->x)+t->orient+ntriabove[t->depth]


void leftChild(TriIndex *t, TriIndex *lt)
{
  lt->depth = t->depth+1;
  lt->orient = lorient[t->orient];
  if (t->depth&1) {
    lt->x = t->x;
    lt->y = t->y;
    lt->s = S0(lt);
  } else {
    lt->x = t->x<<1 | lxinc[t->orient];
    lt->y = t->y<<1 | lyinc[t->orient];
    if (lt->x > xmax[lt->depth] || lt->y > ymax[lt->depth])
      lt->s = 0;
    else
      lt->s = S1(lt);
  }
}

void rightChild(TriIndex *t, TriIndex *rt)
{
  rt->depth = t->depth+1;
  rt->orient = rorient[t->orient];
  if (t->depth&1) {
    rt->x = t->x;
    rt->y = t->y;
    rt->s = S0(rt);
  } else {
    rt->x = t->x<<1 | rxinc[t->orient];
    rt->y = t->y<<1 | ryinc[t->orient];
    if (rt->x > xmax[rt->depth] || rt->y > ymax[rt->depth])
      rt->s = 0;
    else
      rt->s = S1(rt);
  }
}

static void parent(TriIndex *t, TriIndex *p)
{
  p->depth = t->depth-1;
  if (t->depth&1) {
    p->orient = porient1[((t->x&1)<<2)|(t->orient&3)];
    p->x = t->x>>1;
    p->y = t->y>>1;
    p->s = S0(p);
  } else {
    p->orient = porient0[(((t->x^t->y)&1)<<2)|(t->orient&3)];
    p->x = t->x;
    p->y = t->y;
    p->s = S1(p);
  }
}

static void bro1(TriIndex *t, TriIndex *b1)
{
  b1->depth = t->depth;
  b1->x = t->x + b1xinc[t->orient];
  b1->y = t->y + b1yinc[t->orient];
  b1->orient = b1orient[t->orient];
  if (b1->x > xmax[t->depth] || b1->y > ymax[t->depth] ||
      b1->x < 0 || b1->y < 0)
    b1->s = 0;			/* set b1 to be OUTOFBOUNDS */
  else {
    if (t->depth&1)
      b1->s = S1(b1);
    else
      b1->s = S0(b1);
  }
}

static void bro2(TriIndex *t, TriIndex *b2)
{
  b2->depth = t->depth;
  b2->x = t->x + b2xinc[t->orient];
  b2->y = t->y + b2yinc[t->orient];
  b2->orient = b2orient[t->orient];
  if (b2->x > xmax[t->depth] || b2->y > ymax[t->depth] ||
      b2->x < 0 || b2->y < 0)
    b2->s = 0;			/* set b2 to be OUTOFBOUNDS */
  else {
    if (t->depth&1)
      b2->s = S1(b2);
    else
      b2->s = S0(b2);
  }
}

static void bro3(TriIndex *t, TriIndex *b3)
{
  b3->x = t->x+b3xinc[t->orient];
  b3->y = t->y+b3yinc[t->orient];
  b3->orient = b3orient[t->orient];
  b3->depth = t->depth;
  if (b3->x > xmax[t->depth] || b3->y > ymax[t->depth] ||
      b3->x < 0 || b3->y < 0)
    b3->s = 0;			/* set b3 to be OUTOFBOUNDS */
  else {
    if (t->depth&1)
      b3->s = S1(b3);
    else
      b3->s = S0(b3);
  }
}


void nbr1(Hier *hp, TriIndex *t, TriIndex *n1)
{
  TriIndex b1;

  bro1(t,&b1);

  if (b1.s && SMALLN1(hp,t)) {
    leftChild(&b1,n1);
  } else {
    TRIEQUATE(n1,&b1);
  }
}

void nbr2(Hier *hp, TriIndex *t, TriIndex *n2)
{
  TriIndex b2;

  bro2(t,&b2);

  if (b2.s && SMALLN2(hp,t)) {
    rightChild(&b2,n2);
  } else {
    TRIEQUATE(n2,&b2);
  }
}

void nbr3(Hier *hp, TriIndex *t, TriIndex *n3)
{
  TriIndex b3;

  bro3(t,&b3);

  if (b3.s && BIGN3(hp,t)) {
    parent(&b3,n3);
  } else {
    TRIEQUATE(n3,&b3);
  }
}


static int hierDepth;

int init_ntriabove(int depth)
/* initialize ntriabove[] and return size of needed physical address space */
{
  int d,xm,ym;

  /* initialize special Tri's */
  rootL.x=0; rootL.y=0; rootL.depth=1; rootL.orient=NW; rootL.s=1;
  rootR.x=0; rootR.y=0; rootR.depth=1; rootR.orient=SE; rootR.s=2;

  xm = gemXmax-2;	/* count squares (starting at (0,0)) */
  ym = gemYmax-2;
  d = hierDepth = depth;
  while (d >= 1) {
    xmax[d] = xm;
    ymax[d] = ym;
    xm >>= 1;
    ym >>= 1;
    d--;
    xmax[d] = xm;
    ymax[d] = ym;
    d--;
  }

  ntriabove[0] = 0;
  ntriabove[1] = 1;
  d = 1;
  while(d <= depth) {

    /* odd depth: NW,SW,NE,SE triangles */
    ntriabove[d+1]=2*(xmax[d]+1)*(ymax[d]+1)+ntriabove[d];
    d++;

    if (d>depth) break;

    /* even depth: N,S,E,W triangles */
    ntriabove[d+1] = 4*(xmax[d]+1)*(ymax[d]+1)+ntriabove[d];
    d++;
  }
  return(ntriabove[depth+1]+1);/* loop went far enough to calc. this */
}



void getVerts(TriIndex *t, Coord2d p1, Coord2d p2, Coord2d p3)
{
  Coord2d swt,net;
  int d;


  d = (hierDepth-t->depth+1)>>1;
  swt[0] = t->x << d;
  swt[1] = t->y << d;
  net[0] = (t->x+1) << d;
  net[1] = (t->y+1) << d;

  if ((t->depth&1)==0) {
    midpoint(swt,net,p3);
    switch(t->orient) {
    case NORTH:
      p1[0]=net[0]; p1[1]=net[1];
      p2[0]=swt[0]; p2[1]=net[1];
      break;
    case SOUTH:
      p1[0]=swt[0]; p1[1]=swt[1];
      p2[0]=net[0]; p2[1]=swt[1];
      break;
    case EAST:
      p1[0]=net[0]; p1[1]=swt[1];
      p2[0]=net[0]; p2[1]=net[1];
      break;
    case WEST:
      p1[0]=swt[0]; p1[1]=net[1];
      p2[0]=swt[0]; p2[1]=swt[1];
      break;
    default:
      assert(0);
      break;
    }
  } else {
    switch(t->orient) {
    case NW:
      p1[0]=swt[0]; p1[1]=swt[1];
      p2[0]=net[0]; p2[1]=net[1];
      p3[0]=swt[0]; p3[1]=net[1];
      break;
    case SW:
      p1[0]=net[0]; p1[1]=swt[1];
      p2[0]=swt[0]; p2[1]=net[1];
      p3[0]=swt[0]; p3[1]=swt[1];
      break;
    case SE:
      p1[0]=net[0]; p1[1]=net[1];
      p2[0]=swt[0]; p2[1]=swt[1];
      p3[0]=net[0]; p3[1]=swt[1];
      break;
    case NE:
      p1[0]=swt[0]; p1[1]=net[1];
      p2[0]=net[0]; p2[1]=swt[1];
      p3[0]=net[0]; p3[1]=net[1];
      break;
    default:
      assert(0);
      break;
    }
  }
  /*
  printf("(%d,%d,%d,%d) ((%d,%d),(%d,%d),(%d,%d))\n",t->x,t->y,
	 t->depth,t->orient,p1[0],p1[1],p2[0],p2[1],p3[0],p3[1]);
	 */
}
