/*  dem.h -- structures and prototype for reading DEM files  */

#ifndef DEM_H
#define DEM_H

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include "basic.h"
#include "config.h"

/*
 * ============== USGS DEM format ==============
 */
#define MAXROWS 1420		/* the biggest DEM size I've seen */
#define MAXCOLS 1201

typedef struct {
    int rows, cols;
    double dz2m;		/* scale from data units to meters */
    int dzoffset;		/* add this to data item then scale */
    unsigned short dzmax;	/* data item maximum (always >= 0) */

/* information from "A" record */
    char name[145];
    int level;
    int elevation_pattern;
    int reference_system;
    int zone;					/* always zero */
    double map_projection_parameter[15];	/* always zero */
    int ground_unit;
    int elevation_unit;
    int n_polygon;
    UV corner[4];
    double minElevation, maxElevation;
    double axis_angle;
    int accuracy;
    double resolution[3];
    double cellx, celly;

/* information from "B" records */
    int rowsInProfile[MAXCOLS], colsInProfile[MAXCOLS];
    UV firstUVInProfile[MAXCOLS];
    double firstElevationInProfile[MAXCOLS];
    double minElevationInProfile[MAXCOLS], maxElevationInProfile[MAXCOLS];
    unsigned short data[MAXCOLS][MAXROWS];

/* information from "C" record */
    int C2;
    int C2x, C2y, C2z;
    int C2sample_size;
    int C5;
    int C5x, C5y, C5z;
    int C5sample_size;
} Dem;

/*
 * ============== Elevation Model =================
 */
typedef struct {
    int width, height;
    double dz2m;
    int dzoffset;
    unsigned short dzmax;
    Transform *xy2ll;
    UV swUV, nwUV, neUV, seUV;	/* longitude,latitude of 4 corners */
    UV swXY, nwXY, neXY, seXY;	/* x,y of 4 corners */
    unsigned short *data;
} Em;

#define EMDATA(em, i, j) ((em)->data[(int)(j)*(em)->width + (int)(i)])


/* ========== prototypes ================ */

/* dem.c */
Em *readEM(char *filename);
Dem *readDEM(char *filename);



#endif	/* DEM_H */
